#copyright ReportLab Inc. 2001
#see license.txt for license details
#history www.reportlab.co.uk/rl-cgi/viewcvs.cgi/rlextra/graphics/Csrc/renderPM/renderP.py
__version__=''' $Id: renderSWF.py,v 1.5 2003/10/16 16:21:22 admin Exp $ '''
"""Usage:
    from reportlab.graphics import renderSWF
    renderSWF.drawToString(drawing)
"""

from reportlab.pdfbase import _fontdata
from reportlab.graphics.shapes import *
from reportlab.graphics.renderbase import StateTracker, getStateDelta
from reportlab.lib.utils import getStringIO
from reportlab import rl_config

import string, os, sys, tempfile

from types import TupleType, ListType
_SeqTypes = (TupleType,ListType)


# the main entry point for users...
def draw(drawing, canvas):
    R = _SWFRenderer()
    R.draw(drawing, canvas)


def Color2rgb(color):
    (r,g,b) = color.rgb()
    return (r*0xff,g*0xff,b*0xff)


from reportlab.graphics.renderbase import Renderer
class _SWFRenderer(Renderer):

    def __init__(self):
        self._tracker = StateTracker()

    def pop(self):
        self._tracker.pop()
        s = self._tracker.getState()
        self.applyStateChanges(s, {})

    def push(self,node):
        deltas = getStateDelta(node)
        self._tracker.push(deltas)
        self.applyStateChanges(deltas, {})

    def applyStateChanges(self, delta, newState):
        """This takes a set of states, and outputs the PDF operators
        needed to set those properties"""
        for key, value in delta.items():
            if key == 'transform':
                self._canvas.transform(value[0], value[1], value[2],
                                 value[3], value[4], value[5])
            elif key == 'strokeColor':
                self._canvas.setStrokeColor(value)
            elif key == 'strokeWidth':
                self._canvas.setLineWidth(value)
#            elif key == 'strokeLineCap':  #0,1,2
#                self._canvas.setLineCap(value)
#            elif key == 'strokeLineJoin':
#                self._canvas.setLineJoin(value)
#            elif key == 'stroke_dasharray':
#                self._canvas.setDash(array=value)
#            elif key == 'strokeDashArray':
#                if value:
#                    self._canvas.setDash(value)
#                else:
#                    self._canvas.setDash()
            elif key == 'fillColor':
                self._canvas.setFillColor(value)
            elif key in ['fontSize', 'fontName']:
                # both need setting together in PDF
                # one or both might be in the deltas,
                # so need to get whichever is missing
                fontname = delta.get('fontName', self._canvas._fontname)
                fontsize = delta.get('fontSize', self._canvas._fontsize)
                self._canvas.setFont(fontname, fontsize)

    def draw(self, drawing, canvas, x=0, y=0, showBoundary=rl_config.showBoundary):
        """This is the top level function, which
        draws the drawing at the given location.
        The recursive part is handled by drawNode."""
        #stash references for ease of  communication
        self._canvas = canvas
        canvas.__dict__['_drawing'] = self._drawing = drawing
        try:
            if showBoundary: canvas.rect(x, y, drawing.width, drawing.height)
            canvas.saveState()
            deltas = STATE_DEFAULTS.copy()
            deltas['transform'] = [1,0,0,1,x,y]
            self._tracker.push(deltas)
            self.applyStateChanges(deltas, {})
            self.drawNode(drawing)
            self.pop()
            canvas.restoreState()
        finally:
            #remove any circular references
            del self._canvas, self._drawing, canvas._drawing

    def drawNode(self, node):
        """This is the recursive method called for each node
        in the tree"""

        #apply state changes
        self.push(node)

        #draw the object, or recurse
        self.drawNodeDispatcher(node)

        # restore the state
        self.pop()

    def drawGroup(self, group):
        self._canvas.startGroup()
        a, b, c, d, e, f = self._tracker.getCTM()
        for childNode in group.getContents():
            if isinstance(childNode, UserNode):
                node2 = childNode.provideNode()
            else:
                node2 = childNode
            self.drawNode(node2)
        self._canvas.transform(a, b, c, d, e, f)
        self._canvas.endGroup()

    def drawRect(self, rect):
        c = self._canvas
        if rect.rx == rect.ry == 0:
            c.pathBegin()
            c.fillstrokepath()
            c.moveTo(rect.x, rect.y)
            c.lineTo(rect.x+rect.width, rect.y)
            c.lineTo(rect.x+rect.width, rect.y + rect.height)
            c.lineTo(rect.x, rect.y + rect.height)
            c.pathClose()
        else:
            c.roundRect(rect)

    def drawLine(self, line):
        c = self._canvas
        if c.strokeColor is not None:
            c = self._canvas
            c.pathBegin()
            c.pathStroke()
            c.moveTo(line.x1,line.y1)
            c.lineTo(line.x2,line.y2)
            c.pathEnd()

    def drawImage(self, image): pass

    def drawCircle(self, circle):
        c = self._canvas
        c.circle(circle.cx, circle.cy, circle.r)

    def drawPolyLine(self, polyline, _doClose=0):
        P = polyline.points
        assert len(P) >= 2, 'Polyline must have 1 or more points'
        c = self._canvas
        c.pathBegin()
        if _doClose:
            c.pathFill()
        c.pathStroke()
        c.moveTo(P[0], P[1])
        for i in range(2, len(P), 2):
            c.lineTo(P[i], P[i+1])
        if _doClose:
            c.pathClose()
        else:
            c.pathEnd()

    def drawWedge(self, wedge):
        centerx, centery, radius, startangledegrees, endangledegrees = \
         wedge.centerx, wedge.centery, wedge.radius, wedge.startangledegrees, wedge.endangledegrees
        yradius = wedge.yradius
        (x1, y1) = (centerx-radius, centery-yradius)
        (x2, y2) = (centerx+radius, centery+yradius)
        extent = endangledegrees - startangledegrees
        self._canvas.drawArc(x1, y1, x2, y2, startangledegrees, extent, fromcenter=1)

    def drawEllipse(self, ellipse):
        c=self._canvas
        c.ellipse(ellipse.cx, ellipse.cy, ellipse.rx,ellipse.ry)

    def drawPolygon(self, polygon):
        self.drawPolyLine(polygon,_doClose=1)

    def drawString(self, stringObj):
        fill = self._canvas.fillColor
        if fill is not None:
            c = self._canvas
            S = self._tracker.getState()
            text_anchor, x, y, text = S['textAnchor'], stringObj.x, stringObj.y, stringObj.text
            if not text_anchor in ['start','inherited']:
                font, font_size = S['fontName'], S['fontSize']
                textLen = c.getStringWidth(text, font, font_size)
                if text_anchor=='end':
                    x = x-textLen
                elif text_anchor=='middle':
                    x = x - textLen/2
                else:
                    raise ValueError, 'bad value for textAnchor '+str(text_anchor)
            c.drawString(x,y,text)

    def drawPath(self, path):
        from reportlab.graphics.shapes import _renderPath
        c = self._canvas
        drawFuncs = (c.moveTo, c.lineTo, c.curveTo, c.pathClose)
        isClosed = _renderPath(path, drawFuncs)
        if not isClosed:
            c.pathEnd()


from ming import *
class SWFCanvas:
    def __init__(self, width, height):
        self._fonts = {}
        self._shape = None
        self._x = None
        self._y = None
        self.open(width, height)

    def open(self, width, height):
        self._movie = SWFMovie()
        self._movie.setDimension(width+1, height+1)
        self._movie.setRate(12.0)
        self.setFont('Helvetica', 10)
        self._open = 1

    def close(self):
        if self._open:
	    self._movie.nextFrame()
            self._open = 0

    def saveToFile(self,fn):
        self.close()
        return self._movie.save(fn)

    def saveToString(self):
        filename = tempfile.mktemp('.swf')
        self.saveToFile(filename)
        file = open(filename,'rb')
        s = file.read()
        file.close()
        os.unlink(filename)
        return s

    def getStringWidth(self, text, font, size):
        self.setFont(font, size)
        t = SWFText()
        t.setFont(self.font)
        t.setHeight(size)
        t.setSpacing(0)
        return t.getWidth(text)

    def fillstrokepath(self):
        self.pathFill()
        self.pathStroke()

    ### groups ###

    def startGroup(self):
        self.objs = []

    def endGroup(self):
        (a,b,c,d,x,y) = self._transform

        for o in self.objs:
            di = self._movie.add(o)
	    di.setMatrix(a, -b, -c, d, x, self._drawing.height-y)

        self.objs = []

    def addToGroup(self,obj):
        self.objs.append(obj)

    def transform(self, a, b, c, d, x, y):
        self._transform = (a,b,c,d,x,y)

    ### paths ###

    def pathBegin(self):
        self._shape = SWFShape()
        self._x = None
        self._y = None

    def pathClose(self):
        self.lineTo(self._x, self._y)
        self.pathEnd()

    def pathEnd(self):
        if self._shape:
            self._movie.add(self._shape)
        self._shape = None

    def pathFill(self):
        if self.fillColor != None:
            (r,g,b) = Color2rgb(self.fillColor)
            s = self._shape
            s.setLeftFill(s.addFill(r,g,b))

    def pathStroke(self):
        if self.strokeColor != None:
            (r,g,b) = Color2rgb(self.strokeColor)
            s = self._shape
            s.setLine(self.strokeWidth,r,g,b)

    def moveTo(self,x,y):
        if self._shape is None:
            self.pathBegin()
            self.fillstrokepath()
        self._shape.movePenTo(x, self._drawing.height-y)
        self._x = x
        self._y = y

    def lineTo(self,x,y):
        self._shape.drawLineTo(x, self._drawing.height-y)

    def curveTo(self,x1,y1,x2,y2,x3,y3): pass

    def drawArc(self, x1,y1, x2,y2, startAng=0, extent=360, fromcenter=0):
        """Draw a partial ellipse inscribed within the rectangle x1,y1,x2,y2.

        Starting at startAng degrees and covering extent degrees. Angles
        start with 0 to the right (+x) and increase counter-clockwise.
        These should have x1<x2 and y1<y2.
        """

	(angle, arc) = (startAng*pi/180,  (startAng+extent)*pi/180)

        x, y = (x1+x2)/2.0, (y1+y2)/2.0
        rx, ry = (x2-x1)/2.0, (y2-y1)/2.0
        mx = rx * cos(angle) + x
        my = ry * sin(angle) + y
        ax = rx * cos(arc) + x
        ay = ry * sin(arc) + y

        if fromcenter:
            self.moveTo(x,y)
            self.lineTo(mx,my)
        else:
            self.moveTo(mx,my)

        steps = 8
        theta = (arc-angle)/steps
        h = self._drawing.height
        for i in range(0,steps):
                angle += theta
                angleMid = angle-(theta/2)
                cx = x+cos(angleMid)*rx
                cy = y+sin(angleMid)*ry
                dx = x+cos(angle)*rx
                dy = y+sin(angle)*ry
                self.lineTo(cx, cy)
                self.lineTo(dx, dy)

        self.lineTo(ax,ay)

        if fromcenter:
            self.pathClose()
        else:
            self.pathEnd()


#    def drawArc(self, x1,y1, x2,y2, startAng=0, extent=360, fromcenter=0):
#        """Draw a partial ellipse inscribed within the rectangle x1,y1,x2,y2.
#
#        Starting at startAng degrees and covering extent degrees. Angles
#        start with 0 to the right (+x) and increase counter-clockwise.
#        These should have x1<x2 and y1<y2.
#        """
#
#        cx, cy = (x1+x2)/2.0, (y1+y2)/2.0
#        rx, ry = (x2-x1)/2.0, (y2-y1)/2.0
#        mx = rx * cos(startAng*pi/180) + cx
#        my = ry * sin(startAng*pi/180) + cy
#        ax = rx * cos((startAng+extent)*pi/180) + cx
#        ay = ry * sin((startAng+extent)*pi/180) + cy
#
#        str = ''
#        if fromcenter:
#            str = str + "M %f, %f L %f, %f " % (cx, cy, ax, ay)
#
#        if fromcenter:
#            str = str + "A %f, %f %d %d %d %f, %f " % \
#              (rx, ry, 0, extent>=180, 0, mx, my)
#        else:
#            str = str + "M %f, %f A %f, %f %d %d %d %f, %f Z " % \
#              (mx, my, rx, ry, 0, extent>=180, 0, mx, my)
#
#        if fromcenter:
#            str = str + "L %f, %f Z " % (cx, cy)

    def circle(self,x,y,r):
        self.pathBegin()
        self.fillstrokepath()
        self.moveTo(x,y)
        self._shape.drawCircle(r)
        self.pathEnd()

    def ellipse(self,x,y,rx,ry):
	(angle, arc) = (0,  pi*2)
	steps = 8
	theta = arc/steps;
	(cRx, cRy) = (rx/cos(pi/steps), ry/cos(pi/steps))
        self.pathBegin()
        self.fillstrokepath()
	self.moveTo(x+rx, y)
        h = self._drawing.height
	for i in range(0,steps):
		angle += theta
		angleMid = angle-(theta/2)
		cx = x+cos(angleMid)*cRx
		cy = y+sin(angleMid)*-cRy
		dx = x+cos(angle)*rx
		dy = y+sin(angle)*-ry
		self._shape.drawCurveTo(cx, h-cy, dx, h-dy)
        self.pathEnd()

    ### others ###

    def drawString(self,x,y,text):
        (r,g,b) = Color2rgb(self.fillColor)
        t = SWFText()
	t.setFont(self.font)
        t.setHeight(self.font_size)
        t.setColor(r, g, b)
	t.setSpacing(0)
        t.moveTo(x,-y)
        t.addString(text, None)
        self.addToGroup(t)

    ### setFunctions ###

    def setFillColor(self, color):
        self.fillColor = color

    def setStrokeColor(self, color):
        self.strokeColor = color

    def setLineWidth(self, width):
        self.strokeWidth = width

    def setFont(self, font, font_size):
        if self._fonts.has_key(font):
            _font = self._fonts[font]
        else:
            font_file = _fontdata.findT1File(font, '.fdb')
            if font_file is None:
                font_file = _fontdata.findT1File('Helvetica', '.fdb')
            _font = SWFFont(font_file)
            self._fonts[font] = _font

        self._fontname = font
        self._fontsize = font_size
        self.font = _font
        self.font_size = font_size

    def saveState(self): pass

    restoreState = saveState


def drawToSWFCanvas(d):
    c = SWFCanvas(d.width, d.height)
    draw(d, c)
    return c

def drawToFile(d,fn):
    c = drawToSWFCanvas(d)
    c.saveToFile(fn)

def drawToString(d):
    c = drawToSWFCanvas(d)
    return c.saveToString()

save = drawToString
